(***********************************************************************
 *
 * mdMathEvaluator.pas version 1.1
 *
 * Copyright (c) 2016 Michel Deslierres
 * All rights reserved. Tous droits réservés.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 *************************************************************************
 *
 * Changements / Changes
 *
 * - version 1.1
 *
 *     Modified to handle exponentiation, integer division and modulo.
 *
 *************************************************************************)

{: Unit that performs lexical analysis of simple mathematical
   expressions and evaluates them.}
unit mdMathEvaluator;

{$IFDEF FPC}
  {$mode objfpc}
  {$h+ }
{$ENDIF}

{$I parser.inc}

interface

uses
  SysUtils, Math, MiniMath, ParserTypes, mdMathTokenizer;

{$IFDEF UNIT_INFO}
const
  SVersionName = 'mdMathEvaluator';
  SVersion = '1.1';
  VersionMajor = 1;
  VersionMinor = 1;
{$ENDIF}


type
(*:
   Object that performs lexical analysis of simple mathematical
   expressions and evaluates them. This is accomplished in a
   one pass recursive scanner.

   The only supported operators and their precedence are

     op   operation       precedence
     --   --------------  ----------
     ()   parenthesis        5  (also optionally [] {})
     -    negation           4
     ^    exponentiation     3  ( also ** )
     *    multiplication     2
     /    division           2
     %    modulus            2
     :    integer division   2
     +    addition           1
     -    subtraction        1

  WSN definitions of the supported syntax
    -------------------------------------
    Recall:
      a | b  means a or b
      [ a ]  means a is optional              - ex. [a]b means b | ab
      { a }  means 0 or more repetitions of a - ex. {a}b means b | ab | aab | aaab ...
      (   )  groups contained elements        - ex. (a|b)c means ac | bc
    -------------------------------------

DecSym  = ? locale dependant character, usually "." in English, "," in French... ?

Digit           = "0" | ... | "9".
Number          = Digit{Digit} [ DecSym {Digit} [ ("E" | "e") ["-" | "+"] Digit{Digit } ]].
{$IFDEF EXPONENT_LEFT_ASSOCIATIVE}
Exponent   = Factor { ("^" | "**") Factor }.
{$ELSE}
Exponent   = Factor { ("^" | "**") Exponent }.   << default
{$ENDIF}
Term            = Exponent { ("*" | "/" | ":" | "%") Exponent }.
Expression      = Term { ("-" | "+") Term }.
Factor          = Number | "(" Expression  ")" | ("-" | "+") Factor.
*)
  TMathEvaluator = class(TMathTokenizer)
  private
    {$IFnDEF NO_EXCEPTION_MASK}
    FExceptionMask: TFPUExceptionMask;
    FOldExceptionMask: TFPUExceptionMask;
    {: Ensures that the exPrecision bit is always set in the FPU
        exception mask when the LCL is using the Gtk2 widget set.}
    procedure SetExceptionMask(value: TFPUExceptionMask);
    {$ENDIF}
    function Factor: float;
    function Exponent: float;
    function Expression: float;
    function Term: float;

  public
    {$IFnDEF NO_EXCEPTION_MASK}
    constructor create;
    {$ENDIF}

    {: The function that performs a lexical analysis of a mathematical expression
       and returns the result of the calculation:

           EvaluateExpression('2/8') --> 0.25

       Reports errors when the expression is not well formed or when a run-time
       error such as division by zero occurs. The properties ErrorCode and
       ErrorMessage are updated before returning from the function. In addition,
       if the boolean property RaiseExceptions is true an exception of type
       EParseError (see ParserTypes) is raised.

       Which of the mathematical run-time errors are reported depends on the
       FPU exception mask. If the SET_EXCEPTION_MASK directive is defined
       in parser.inc, then the desired FPU exception mask must be set
       in the ExceptionMask property. If the directive is not defined, then
       use the function SetExceptionMask from Math.pas. }
    function EvaluateExpression(const aSource: string): float;

    {$IFnDEF NO_EXCEPTION_MASK}
    {: FPU exception mask that will be used during each call to EvaluateExpression.
       The previous exceptions mask is restored as soon as possible after
       evaluation including before invoking any exception handler. The default
       value is [exDenormalized, exUnderflow, exPrecision] which corresponds to
       the default value set by Free Pascal's rtl. On most x86_64 platforms that
       will be the 'system' exception mask but the gtk2 widget set does mask all
       exceptions (this is done in the TGtk2WidgetSet.AppInit procedure
       invoked in the Application.Initialize procedure at the startup of a
       graphic application using the LCL).}
    property ExceptionMask: TFPUExceptionMask read FExceptionMask write SetExceptionMask;
      //default [exDenormalized, exUnderflow, exPrecision]
    {$ENDIF}
  end;

implementation

uses
  ParserConsts;

{$IFnDEF NO_EXCEPTION_MASK}
constructor TMathEvaluator.Create;
begin
  inherited;
  FExceptionMask := [exDenormalized, exUnderflow, exPrecision];
end;

procedure TMathEvaluator.SetExceptionMask(value: TFPUExceptionMask);
begin
  FExceptionMask := value {$IFDEF LCLGTK2} + [exPrecision] {$ENDIF};
end;
{$ENDIF}

function TMathEvaluator.Exponent: float;
begin
  result := Factor;
  while TokenType = ttCaret do begin
    NextTrueToken;
    {$IFDEF EXPONENT_LEFT_ASSOCIATIVE}
    result := Power(result, factor)
    {$ELSE}
    result := Power(result, Exponent());  // don't forget () in fpc mode
    {$ENDIF}
  end;
end;

function TMathEvaluator.Term: float;
begin
  result := Exponent;
  repeat
    if TokenType = ttMult then begin
      NextTrueToken;
      result := result * Exponent;
    end
    else if TokenType = ttDiv then begin
      NextTrueToken;
      result := result / Exponent;
    end
    else if TokenType = ttColon then begin
      NextTrueToken;
      result := IDiv(result, Exponent);
    end
    else if TokenType = ttPercent then begin
      NextTrueToken;
      result := IMod(result, Exponent);
    end
   else begin
      exit;
    end;
  until false;
end;

function TMathEvaluator.Expression: float;
begin
  result := Term;
  repeat
    if TokenType = ttPlus then begin
      NextTrueToken;
      result := result + Term;
    end
    else if TokenType = ttMinus then begin
      NextTrueToken;
      result := result - Term;
    end
    else begin
      exit;
    end;
  until false;
end;

function TMathEvaluator.Factor: float;
var
  wantedRightParenthesis: char;
begin
  result := 0;
  if TokenType = ttNumber then begin
    result := sfloat(Token);
    NextTrueToken;
  end
  else if TokenType = ttPlus then begin
    NextTrueToken;  // skip and ignore unary '+'
    result := Factor(); // don't forget ()
  end
  else if TokenType = ttMinus then begin
    NextTrueToken;  // skip unary '-'; return negative of following factor
    result := - Factor(); // don't forget ()
  end
  else if TokenType = ttLeftParenthesis then begin
    case Token[1] of
      '(': wantedRightParenthesis := ')';
      '[': wantedRightParenthesis := ']';
      '{': wantedRightParenthesis := '}';
    end;
    NextTrueToken;  // skip '('
    result := Expression;
    if (TokenType <> ttRightParenthesis) or (Token <> wantedRightParenthesis) then begin
      ParserError(meExpectedRightPar);
    end;
    NextTrueToken; // skip ')'
  end
  else if TokenType = ttEOS then
    ParserError(meUnexpectedEOS)
  else
    ParserError(meUnexpectedElement);
end;

function TMathEvaluator.EvaluateExpression(const aSource: string): float;
begin
  Source := aSource; // resets error state
  {$IFnDEF NO_EXCEPTION_MASK}
  FOldExceptionMask := Math.SetExceptionMask(FExceptionMask);
  {$ENDIF}
  NextTrueToken;     // prime the pump
  try
    result := Expression;
    if (TokenType <> ttEOS) then
      // some extra bits found in the source, signal error
      ParserError(meExpectedEOS);
  except
    // exception handling - only pass on exception if RaiseExceptions is true
    // converting all general (non parsing exceptions) exceptions to
    // type EParserError. Restore the fpu exception mask before
    // passing on the exception.
    On E: EParserError do begin
      if RaiseExceptions then begin
        {$IFnDEF NO_EXCEPTION_MASK}
        Math.SetExceptionMask(FOldExceptionMask);
        {$ENDIF}
        Raise;
      end;
    end;
    On E: Exception do begin
      SelectAll; // select all the source
      {$IFnDEF NO_EXCEPTION_MASK}
      Math.SetExceptionMask(FOldExceptionMask);
      {$ENDIF}
      // distinguish math exceptions from other exceptions and let
      // the Error method re raise exception or not according to the
      // RaiseException property.
      if (E is EMathError) or (E is EDivByZero) then
        Error(meEvalError, Format(SmeEvalError, [lowercase(E.Message)]))
      else
        Error(meOtherError, Format(SmeOtherError, [lowercase(E.Message)]));
    end;
  end;
  {$IFnDEF NO_EXCEPTION_MASK}
  // normal exit after parsing and evaluating an expression without error;
  // restore the privious exception mask.
  Math.SetExceptionMask(FOldExceptionMask);
  {$ENDIF}
end;

end.
