(***********************************************************************
 *
 * mdMathTokenizer.pas version 1.1
 *
 * Copyright (c) 2013-2016 Michel Deslierres
 * All rights reserved. Tous droits réservés.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier licence.txt
 *
 *************************************************************************
 *
 * Formerly called SimpleTokenizer.pas this is a Free Pascal version.
 *
 * *** LINUX NOTE ***
 *
 * Add the unit clocale at the start of the uses clause of the
 * project file ( *.lpr ). That way, DefaultFormatSettings and
 * hence the decimal symbol are update to the user's locale
 * at program startup. Otherwise, the DefaultFormatSettings will
 * be the US settings no matter the user's locale.
 *
 *************************************************************************
 *
 * Changements / Changes
 *
 * - version 1.1
 *
 *     Tokenizer modified to recognize exponentiation operators '^' and
 *     '**', the integer division and modulus operators ':' and '%'and 
 *     the comment operator '//'. Also added two other pairs of
 *     parenthesis [] and {}. Left and right parenthesis must match
 *      [ (.. ) ] is ok but [ ( .. ) ) is not!
 *
 *************************************************************************)


{: Unit that defines a tokenizer object that is used by the parsers
   to get individual parts of a mathematical expressions.}
unit mdMathTokenizer;

{$IFDEF FPC}
  {$mode objfpc}
  {$h+ }
{$ENDIF}  

{$I parser.inc}

interface

uses
  SysUtils, ParserTypes;

{$IFDEF UNIT_INFO}
const
  SVersionName = 'mdMathTokenizer';
  SVersion = '1.1';
  VersionMajor = 1;
  VersionMinor = 1;
{$ENDIF}

type
    {: Basic tokenizer. This object scans a source (usually a single line) 
       and separates it into meaningfull strings assigning them a type. }
  TMathTokenizer = class
  private
    FSource: string;
    FFormatSettings : TFormatSettings;
    FTokenType: TMathTokenType;
    FErrorCode: TMathError;
    FErrorMessage: string;
    FRaiseExceptions: boolean;
    FTokenStart: integer;   // Index of the current token's position
    FTokenLength: integer;  // The length of the current token.
    FNextCharPos: integer;  // Index of the character in source after the current token
    FLeftBraces: string;
    FRightBraces: string;
    function GetToken: string;
    procedure SetSource(const value: string);
  protected
      {: Raises an EMathError with the specified error code and error
         message.}
    procedure Error(Code: TMathError; const msg: string);

      {: Raises an EMathError with the specified parser error code. The
         error message is obtained from function ParserErrorMessage in
         ParserConsts.}
    procedure ParserError(Code: TParserError);

      {: Called when parsing a token to increment FNextCharPos and
         FTokenLength by one. }
    procedure MoveByOneChar;

      {: Called when scanning a comment which has the syntax
         <expression> // a comment
         where the comment will be '// a comment' and the
         following end of line marker is ignored.}
    procedure MoveToEndOfLine;

     {: Looks ahead to the next character after the current token. If
         there are no more characters in the source buffer, returns #0.}
    function Peek: char;

      {: Restarts scanning at the beginning of the source. Used in testing
         but there is probably no reason to call this procedure directly.}
    procedure Reset;

      {: Sets TokenStart and TokenLength to contain the complete source
         string. Used when an evaluation error is found.}
    procedure SelectAll;

  public
    constructor Create;

      (*: Add character aValue as an accepted left parenthesis. A matching
          right parenthesis will also be added. Only '{' and '[' can be
          added, all other characters are ignored. Initially these
          are already accepted and do not need to be added. *)
    procedure AddLeftBrace(aValue: char);

      (*: Remove character aValue from the list of accepted left parenthesis.
          The matching right parenthesis will also be removed. Only '{' and
          '[' can be removed, '(' is always accepted as a left parenthesis. *)
    procedure DeleteLeftBrace(aValue: char);

      {: Returns true if the given character c is an allowable decimal
         separator. Usually the decimal symbol is a punctuation mark but
         it cannot be an operator, a digit, a letter nor white space.
         This utility function is not used in the tokenizer and parser.}
    function canBeDecimalSeparator(c: char): boolean;

      {: Returns the source string converting all decimal separators to
         periods ('.') in numbers. Will not change periods found in
         comments. If the directive NO_NLS is defined the function
         returns the source string without change.
         This utility function is not used in the tokenizer and parser.}
    function Encode(const source: string): string;

      {: Returns the source string converting all periods ('.') in numbers
         to the current decimal separator. Will not change periods found in
         comments. If the directive NO_NLS is defined the function
         returns the source string without change.
         This utility function is not used in the tokenizer and parser.}
    function Decode(const source: string): string;

      {: Returns next token's type. Properties TokenStart, TokenLength,
         TokenType and Token are all adjusted. The protected variable
         FNextCharPos is updated to point to the first character in the
         source following the current token. The next call to NextToken
         will start from FNextCharPos.

         If all the source has been scanned then NextToken returns ttEOS
         and
           TokenStart = length(Source) + 1
           TokenLength = 0
           TokenType = ttEOS
           Token = ''.

         This procedure MUST NOT call on Error, ParserError nor raise
         an exception.}
    function NextToken: TMathTokenType; virtual;

      {: Returns the next token that is not white space (of type ttWhite).}
    function NextTrueToken: TMathTokenType;

      {: Returns the value floating point value corresponding to the given
         string using the tokenizer's FormatSettings. Will raise an
         EConvertErr exception if the string is not well formed.
         Used internally by the parser which will be a descendant
         of this class.}
    function sfloat(const value: string): float;

      {: Returns a string representation of the given floating point value
         using the tokenizer's FormatSettings. This utility function is not
         used in the tokenizer and parser.}
    function floats(const value: float): string;

      {: Decimal symbol used by sfloat to convert a string to a real value.
         This is part of the TFormatSettings record used by the tokenizer.
         No error checking is done when setting the value. However see
         the function canBeDecimalSepartor.}
    property DecimalSeparator: char read FFormatSettings.DecimalSeparator
                                   write FFormatSettings.DecimalSeparator;

      {: Local copy of error code contained in EMathError exception.}
    property ErrorCode: TMathError read FErrorCode;

      {: Local copy of the error message contained in EMathError exception.}
    property ErrorMessage: string read FErrorMessage;

      {: Local copy of DefaultFormatSettings that can modified to change
         the decimal point. See DecimalSeparator. This is used by the
         methods sfloat and floats}
    property FormatSettings: TFormatSettings read FFormatSettings write FFormatSettings;

      (*: List of accepted left parenthesis. By default these are '([{'. *)
    property LeftBraces: string read FLeftBraces;

      {: If true, Error raises EParserError exceptions. If false Error only
         updates the ErrorCode and ErrorMessage.}
    property RaiseExceptions: boolean read FRaiseExceptions write FRaiseExceptions;

      (*: List of accepted right parenthesis. By default these are ')]}'. *)
    property RightBraces: string read FRightBraces;

      {: String containing the source to be scanned. Whenever the source
         is set, the protected Reset method is called that resets the
         TokenType, TokenStart, TokenLength, ErrorCode, and ErrorMessage
         properties .}
    property Source: string read FSource write SetSource;

      {: Index into Source of the first character of the current token.
         Equal to length(Source) + 1 if at end of source.}
    property TokenStart: integer read FTokenStart;

      {: Length of the current token, 0 if at end of source.}
    property TokenLength: integer read FTokenLength;

      {: The current token as a string. The tokenizer does not store a
         string, but uses TokenStart and TokenLength to retrieve the
         current token from the source.}
    property Token: string read GetToken;

      {: The type of the current token.}
    property TokenType: TMathTokenType read FTokenType;
  end;


{$IFnDEF NO_NLS}

const
  {Default en_US format settings. Used by encode and decode if NO_NLS is
   not defined to write/read numbers. }
  EnFormatSettings : TFormatSettings = (
    CurrencyFormat: 1;
    NegCurrFormat: 5;
    ThousandSeparator: ',';
    DecimalSeparator: '.';
    CurrencyDecimals: 2;
    DateSeparator: '-';
    TimeSeparator: ':';
    ListSeparator: ',';
    CurrencyString: '$';
    ShortDateFormat: 'd/m/y';
    LongDateFormat: 'dd" "mmmm" "yyyy';
    TimeAMString: 'AM';
    TimePMString: 'PM';
    ShortTimeFormat: 'hh:nn';
    LongTimeFormat: 'hh:nn:ss';
    ShortMonthNames: ('Jan','Feb','Mar','Apr','May','Jun',
                      'Jul','Aug','Sep','Oct','Nov','Dec');
    LongMonthNames: ('January','February','March','April','May','June',
                     'July','August','September','October','November','December');
    ShortDayNames: ('Sun','Mon','Tue','Wed','Thu','Fri','Sat');
    LongDayNames:  ('Sunday','Monday','Tuesday','Wednesday','Thursday','Friday','Saturday');
    TwoDigitYearCenturyWindow: 50;);

{$ENDIF}

implementation

uses
  ParserConsts;

const
  CR = #13;
  LF = #10;

{ Utility functions }

function isWhite(c: char): boolean;
begin
  result := c <= ' ';
end;

function isDigit(c: char): boolean;
begin
  result := (c >= '0') and (c <= '9');
end;

function isLetter(c: char): boolean;
begin
  c := upcase(c);
  result := (c >= 'A') and (c <= 'Z')
end;

function isOperator(c: char): boolean;
const
  Operators: string = '()[]{}*+-/%:^';
begin
  result := pos(c, Operators) > 0;
end;

{ TMathTokenizer }

constructor TMathTokenizer.Create;
begin
  inherited Create;
  FLeftBraces := '([{';
  FRightBraces := ')]}';
  FFormatSettings := DefaultFormatSettings;
  // At least one Windows locale, Persian (Iran) (fa_IR), has a slash
  // as decimal separator.
  if DecimalSeparator = '/' then
    DecimalSeparator := '.';
  Reset;
end;

procedure TMathTokenizer.AddLeftBrace(aValue: char);
var
  c: char;
begin
  if (aValue in ['[', '{']) and (pos(AValue, FLeftBraces) < 1) then begin
    FLeftBraces := FLeftBraces + aValue;
    if aValue = '[' then
      c := ']'
    else
      c := '}';
    FRightBraces := FRightBraces + c;
  end;
end;

procedure TMathTokenizer.DeleteLeftBrace(aValue: char);
var
  c: char;
  i: integer;
begin
  i := pos(AValue, FLeftBraces);
  if (i > 0) and (aValue <> '(') then begin
    delete(FLeftBraces, i, 1);
    if aValue = '[' then
      c := ']'
    else
      c := '}';
    i := pos(c, FRightBraces);
    if i > 0 then
      delete(FRightBraces, i, 1);
  end;
end;

function TMathTokenizer.canBeDecimalSeparator(c: char): boolean;
begin
  result := not ( isOperator(c) or isDigit(c) or isLetter(c) or isWhite(c) );
end;

procedure TMathTokenizer.Error(Code: TMathError; const msg: string);
var
  E: EParserError;
begin
  // store local copies of the error code and message
  FErrorCode := code;
  FErrorMessage := msg;
  if RaiseExceptions or (Code in ParserErrors) then begin
    // raise the exception
    E := EParserError.Create(msg);
    E.Error := Code;
    E.TokenStart := TokenStart;
    E.TokenLength := TokenLength;
    Raise E;
  end;
end;

function TMathTokenizer.floats(const value: float): string;
begin
  result := sysutils.floattostr(value, FormatSettings);
end;

function TMathTokenizer.GetToken: string;
begin
  result := copy(FSource, FTokenStart, FTokenLength)
end;

procedure TMathTokenizer.MoveByOneChar;
begin
  inc(FNextCharPos);
  inc(FTokenLength);
end;

{ How are EOL marked ?

  classic DOS, Windows   CR LF
  Unix, Linux            LF
  older Mac OS           CR
  }
procedure TMathTokenizer.MoveToEndOfLine;
begin
  while (FNextCharPos <= length(FSource)) and not (FSource[FNextCharPos] in [LF, CR]) do
    MoveByOneChar;
  while (FNextCharPos <= length(FSource)) and (FSource[FNextCharPos] in [LF, CR]) do
    MoveByOneChar;
end;

function TMathTokenizer.NextToken: TMathTokenType;

  {There are three basic types of tokens: white space, numbers and operators.
   Hence one of the following three procedures is called on the basis of the
   character as FNextCharPos. Each procedure attempts to accumulate as many
   characters in the token as possible given it's type.}

  {Any character equal to or less than as space ' ' is considered a
   white character. See isWhite.}
  procedure GetWhite;
  begin
    result := ttWhite;
    while (FNextCharPos <= length(FSource)) and isWhite(FSource[FNextCharPos]) do
      MoveByOneChar;
  end;

  {No distinction is made between integers and floating numbers. It would
   be easy to replace ttNumber with ttInteger and ttFloat and to differentiate
   between the two while parsing the token as indicated below in comments
   beginning with a single star.

   Another approach could be to keep ttNumber and to add a boolean called
   isFloat for example. This is indicated with two starts}
  procedure GetNumber;
  begin
    result := ttNumber; {* := ttInteger as default} {** isFloat := false}
    {parse leading digits}
    while (FNextCharPos <= length(FSource)) and isDigit(FSource[FNextCharPos]) do
      MoveByOneChar;
    {check if decimal point}
    if (FNextCharPos <= length(FSource)) and (FSource[FNextCharPos] = DecimalSeparator) then
      MoveByOneChar; {* result := ttFloat} {** isFloat := true}
    {parse decimal part if present}
    while (FNextCharPos <= length(FSource)) and isDigit(FSource[FNextCharPos]) do
      MoveByOneChar;
    {parse exponent if present}
    if (FNextCharPos <= length(FSource)) and (upcase(FSource[FNextCharPos]) = 'E') then begin
      MoveByOneChar;
      {parse exponents + or - if present}
      if (FNextCharPos <= length(FSource)) and (FSource[FNextCharPos] in ['-','+']) then
        MoveByOneChar;

      { could enforce need for a digit following E
      if (FNextCharPos > length(FSource)) or not isDigit(FSource[FNextCharPos]) then
        ParserError(meInvalidNumber); }

      {parse exponent if present}
      while (FNextCharPos <= length(FSource)) and isDigit(FSource[FNextCharPos]) do
        MoveByOneChar;
    end;
  end;

  procedure GetOperator(c: char);
  begin
    MoveByOneChar; {skip c}
    case c of
      '%': result := ttPercent;
      '*': if Peek <> '*' then
             result := ttMult
           else begin
             result := ttCaret;
             MoveByOneChar;
           end;
      '+': result := ttPlus;
      '-': result := ttMinus;
      '/': if Peek <> '/' then
             result := ttDiv
           else begin
             result := ttComment;
             MoveToEndOfLine;
           end;
      ':': result := ttColon;
      '^': result := ttCaret;
    else
      begin
        if (pos(c, FLeftBraces) > 0) then
          result := ttLeftParenthesis
        else if (pos(c, FRightBraces) > 0) then
          result := ttRightParenthesis
        else
          result := ttUnknown;
      end;
    end;
  end;

var
  c: char;
begin
  // move past current token
  FTokenStart := FNextCharPos;
  FTokenLength := 0;
  if FNextCharPos > length(FSource) then
    result := ttEOS
  else begin
    c := FSource[FTokenStart];
    if isWhite(c) then
      GetWhite
    else if isDigit(c) then
      GetNumber
    else
      GetOperator(c);
  end;
  FTokenType := result;
end;

function TMathTokenizer.NextTrueToken: TMathTokenType;
begin
  repeat
    result := NextToken;
  until (result <> ttWhite) and (result <> ttComment);
end;

procedure TMathTokenizer.ParserError(Code: TParserError);
begin
  Error(Code, ParserErrorMessage(Code));
end;

function TMathTokenizer.Peek: char;
begin
  if FNextCharPos > length(FSource) then
    result := #0
  else
    result := FSource[FNextCharPos];
end;

procedure TMathTokenizer.Reset;
begin
  FTokenStart := 1;
  FTokenLength := 0;
  FNextCharPos := 1;
  FTokenType := ttUnknown;
  FErrorCode := meNoError;
  FErrorMessage := '';
end;

procedure TMathTokenizer.SelectAll;
begin
  FTokenStart := 1;
  FTokenLength := length(FSource);
end;

procedure TMathTokenizer.SetSource(const value: string);
begin
  FSource := value;
  Reset;
end;

function TMathTokenizer.sfloat(const value: string): float;
begin
  if not trystrtofloat(value, result, FormatSettings) then begin
    result := nan;
    ParserError(meInvalidNumber);
  end;
end;

{$IFnDEF NLS}

function TMathTokenizer.Encode(const source: string): string;
var
  tempTokenizer: TMathTokenizer;
  s: string;
  p: integer;
begin
  if DecimalSeparator = enFormatSettings.DecimalSeparator then begin
    result := source;
    exit;
  end;
  result := '';
  tempTokenizer := TMathTokenizer.create;
  try
    tempTokenizer.FormatSettings := FormatSettings;
    tempTokenizer.Source := source;
    while tempTokenizer.NextToken <> ttEOS do begin
      s := tempTokenizer.Token;
      if tempTokenizer.TokenType = ttNumber then begin
        p := pos(DecimalSeparator, s);
        if p > 0 then
          s[p] := EnFormatSettings.DecimalSeparator;
      end;
      result := result + s;
    end;
  finally
    tempTokenizer.free;
  end;
end;

function TMathTokenizer.Decode(const source: string): string;
var
  tempTokenizer: TMathTokenizer;
  s: string;
  p: integer;
begin
  if DecimalSeparator = enFormatSettings.DecimalSeparator then begin
    result := source;
    exit;
  end;
  result := '';
  tempTokenizer := TMathTokenizer.create;
  try
    tempTokenizer.FormatSettings := enFormatSettings;
    tempTokenizer.Source := source;
    while tempTokenizer.NextToken <> ttEOS do begin
      s := tempTokenizer.Token;
      if tempTokenizer.TokenType = ttNumber then begin
        p := pos(enFormatSettings.DecimalSeparator, s);
        if p > 0 then
          s[p] := DecimalSeparator;
      end;
      result := result + s;
    end;
  finally
    tempTokenizer.free;
  end;
end;

{$ELSE}

function TMathTokenizer.Encode(const source: string): string;
begin
  result := source;
end;

function TMathTokenizer.Decode(const source: string): string;
begin
  result := source;
end;

{$ENDIF}


end.
