(***********************************************************************
 *
 * TestParser.pas
 *
 * Copyright (c) 2016 Michel Deslierres
 * All rights reserved. Tous droits réservés.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 **************************************************************************
 *
 * This tests a lot of errors, it will be a lot faster to run the test
 * outside the Lazarus IDE
 *
 **************************************************************************
 *
 * Version 1.1,
 *
 * Augmented to test IntDiv, Mod, Power and added [] and {}
 * parenthesis.
 *
 *************************************************************************)

unit TestParser;

interface

uses
  Classes, SysUtils, fpcunit, testregistry,
  Math, ParserTypes, mdMathEvaluator;

type
  TestMathEvaluator = class(TTestCase)
  strict private
    FSource: string;
    FEvaluator: TMathEvaluator;
  protected
    // TestOk: added eps & msg in version 1.1
    procedure TestOk(const value: string; wanted: float; eps: float = 0; const msg: string = '');
    procedure TestError(const value: string; error: TMathError);
  public
    procedure SetUp; override;
    procedure TearDown; override;
  published
    procedure TestCreate;
    procedure TestFactor0;
    procedure TestFactor10;
    procedure TestFactorMinus10;
    procedure TestFactorFloat;
    procedure TestInvalidNumber;
    procedure TestTerm1;
    procedure TestTerm2;
    procedure TestExpression1;
    procedure TestExpression2;
    procedure TestExpression3;
    procedure TestExpression4;
    procedure TestExpression5;
    procedure TestExpression6;
    procedure TestMultiline;
    procedure TestMultiline2;
    procedure TestUnexpectedEOS;
    procedure TestExpectedEOS;
    procedure TestUnexpectedElement;
    procedure TestExpecteRightPar;
    procedure TestDivBy0;
    procedure TestOverflow;
    procedure TestUnderflow;
    procedure TestIntDiv;      // added in version 1.1
    procedure TestMod;         // added in version 1.1
    procedure TestPower;       // added in version 1.1
    procedure TestPower2;      // added in version 1.1
    procedure TestPower3;      // added in version 1.1
    procedure TestPower4;      // added in version 1.1
    procedure TestMixedParenthesis;      // added in version 1.1
    procedure TestMismatchedParethensis; // added in version 1.1
    procedure TestComment;               // added in version 1.1
  end;

implementation

procedure TestMathEvaluator.SetUp;
begin
  FSource := '';
  FEvaluator := TMathEvaluator.create;
  FEvaluator.DecimalSeparator := '.';
  FEvaluator.RaiseExceptions := true;
end;

procedure TestMathEvaluator.TearDown;
begin
  FSource := '';
  freeandnil(FEvaluator);
end;

{ old version 1.0
procedure TestMathEvaluator.TestOk(const value: string; wanted: float);
begin
  CheckEquals(wanted, FEvaluator.EvaluateExpression(value));
end;
}

procedure TestMathEvaluator.TestOk(const value: string; wanted, eps: float; const msg: string);
var
  res: float;
begin
  try
    if eps < 0 then begin
      res := FEvaluator.EvaluateExpression(value);
      if abs((res-wanted)/wanted) > -eps then
        CheckTrue(false, Format('Expected %g got %g %s', [wanted, res, msg]))
      else
        CheckTrue(true, 'so DUnit does not complain about missing test');
    end
    else
      CheckEquals(wanted, FEvaluator.EvaluateExpression(value), eps, msg);
  except
    On E: EParserError do begin
      CheckTrue(false, E.Message);
    end;
  end;
end;

procedure TestMathEvaluator.TestError(const value: string; error: TMathError);
begin
  try
    FEvaluator.EvaluateExpression(value);
  except
    On E: EParserError do begin
      CheckEquals(ord(error), ord(E.Error), E.Message);
    end;
  end;
end;

procedure TestMathEvaluator.TestOverflow;
begin
  if sizeof(float) = sizeof(extended) then
    TestError('1.1E4930 * 1.1E4930', meEvalError)
  else if sizeof(float) = sizeof(double) then
    TestError('1.1E307 * 1.1E307', meEvalError)
  else if sizeof(float) = sizeof(single) then
    TestError('1.1E38 * 1.1E38', meEvalError)
  else
    CheckTrue(false, 'Unknown float type');
end;

procedure TestMathEvaluator.TestUnderflow;
begin
  if sizeof(float) = sizeof(extended) then
    TestError('2/1E4930/1E4930', meEvalError)
  else if sizeof(float) = sizeof(double) then
    TestError('2/1E306/1E306', meEvalError)
  else if sizeof(float) = sizeof(single) then
    TestError('2/1E38/1E38', meEvalError)
  else
    CheckTrue(false, 'Unknown float type');
end;

procedure TestMathEvaluator.TestCreate;
begin
  TestError('',  meUnexpectedEOS);
end;

procedure TestMathEvaluator.TestDivBy0;
begin
  TestError('5.8/(2 - 2)', meEvalError);
end;

procedure TestMathEvaluator.TestUnexpectedEOS;
begin
  TestError('5 +', meUnexpectedEOS);
  TestError('6 *', meUnexpectedEOS);
end;

procedure TestMathEvaluator.TestExpectedEOS;
begin
  TestError('5 + 8 32', meExpectedEOS);
  TestError('6 * 8 32', meExpectedEOS);
  TestError('7 * 8 !', meExpectedEOS);
  TestError('8#', meExpectedEOS);
end;

procedure TestMathEvaluator.TestUnexpectedElement;
begin
  TestError('5 + !', meUnExpectedElement);
  TestError('+*', meUnExpectedElement);
end;

procedure TestMathEvaluator.TestExpecteRightPar;
begin
  TestError('5*(2 + 3', meExpectedRightPar);
end;

procedure TestMathEvaluator.TestInvalidNumber;
begin
  TestError('5.3E*8', meInvalidNumber);
  TestError('5.3E3456789', meInvalidNumber);
  TestError('5.3E-3456789', meInvalidNumber);
end;

procedure TestMathEvaluator.TestExpression1;
begin
  TestOk('5 - 32', 5 - 32);
end;

procedure TestMathEvaluator.TestExpression2;
begin
  TestOk('5 - 32 + - 34', 5 - 32 + - 34);
end;

procedure TestMathEvaluator.TestExpression3;
begin
  TestOk('5 - 32 * 4', 5 - 32 * 4);
end;

procedure TestMathEvaluator.TestExpression4;
begin
  TestOk('5 - 32 * - 34', 5 - 32 * - 34);
end;

procedure TestMathEvaluator.TestExpression5;
begin
  TestOk('(5 - 32) * 4', (5 - 32) * 4);
end;

procedure TestMathEvaluator.TestExpression6;
begin
  TestOk('(6 - 2) / (3 - 1)', (6 - 2) / (3 - 1));
end;

procedure TestMathEvaluator.TestFactor0;
begin
  TestOk('0', 0);
end;

procedure TestMathEvaluator.TestFactor10;
begin
  TestOk(' 10 ', 10);
end;

procedure TestMathEvaluator.TestFactorFloat;
begin
  TestOk(' 10.345E12', 10.345E12)
end;

procedure TestMathEvaluator.TestFactorMinus10;
begin
  TestOk(' -  10 ', -10);
end;

procedure TestMathEvaluator.TestMultiline;
begin
  TestOk(' 4 *'#13#10' 32 - '#13#10'5', 4 * 32 - 5);
end;

procedure TestMathEvaluator.TestMultiline2;
begin
  TestOk(' 4 *'#10' 32 - '#10'5', 4 * 32 - 5);
end;

procedure TestMathEvaluator.TestTerm1;
begin
  testOk('32.8 *72.22', 32.8 *72.22);
end;

procedure TestMathEvaluator.TestTerm2;
begin
  testOk('32 * 72 / 8', 32 * 72 / 8);
end;

// added in version 1.1

procedure TestMathEvaluator.TestIntDiv;
begin
  TestOk('9 : 2', 9 div 2);
  TestOK('-9 : 2', -9 div 2);
  TestOK('(-9) : 2', (-9) div 2);
  TestOK('9 : -2', 9 div -2);
  TestOK('-9 : -2', -9 div -2);
end;

procedure TestMathEvaluator.TestMod;
begin
  TestOk('9 % 2', 9 mod 2);
  TestOK('-9 % 2', (-9) mod 2);
  TestOK('(-9) % 2', (-9) mod 2);
  TestOK('9 % -2', 9 mod -2);
  TestOK('-9 % -2', -9 mod -2);
end;


procedure TestMathEvaluator.TestPower;
begin
  TestOk('2^3', power(2, 3));
end;

procedure TestMathEvaluator.TestPower2;
begin
  {$IFDEF USE_MATH_UNIT}
  TestOk('2^3^4', power(2, power(3, 4)));
  {$ELSE}
  TestOk('2^3^4', power(2, power(3, 4)), -1E-16);
  {$ENDIF}
end;

procedure TestMathEvaluator.TestPower3;
begin
  TestOk('5^0', power(5, 0));
  TestOk('5^0', 1);
end;

procedure TestMathEvaluator.TestPower4;
begin
  TestOk('0^0', power(0, 0));
  TestOk('0^0', 1);
end;

procedure TestMathEvaluator.TestMixedParenthesis;
begin
  TestOK('2*(5*{3-1}/[2+2])/9', 2*(5*(3-1)/(2+2))/9);
end;

procedure TestMathEvaluator.TestMismatchedParethensis;
begin
  TestError('2*(5*{3-1]/[2+2])/9', meExpectedRightPar);
end;

procedure TestMathEvaluator.TestComment;
begin
  TestOK('2*//comment'#13#10'5', 10);
end;

initialization
  RegisterTest(TestMathEvaluator);
end.

