(***********************************************************************
 *
 * TestTokenizer.pas
 *
 * Copyright (c) 2016 Michel Deslierres
 * All rights reserved. Tous droits réservés.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 *************************************************************************
 *
 * Version 1.1,
 *
 * Augmented to test new operators  ^, **, %, : and //
 * and Encode and Decode methods. 
 *
 *************************************************************************)

unit TestTokenizer;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, fpcunit,  testregistry,
  ParserTypes, mdMathTokenizer;

type
  TestNextTokenType = class(TTestCase)
  strict private
    FTokenizer: TMathTokenizer;
    FSource: string;
    FSources: TStrings;
  protected
    procedure SetSource;
    procedure TestSource;
  public
    procedure SetUp; override;
    procedure TearDown; override;
  published
    procedure TestCreate;
    procedure TestWhiteChar;   // ' '
    procedure TestWhiteString; // '  '
    procedure TestDigit;       // '0' to '9';
    procedure TestWhiteDigit;  // ' 0 ' to ' 9 ';
    procedure TestDoubleDigit; // '10' to '19';
    procedure TestDecimals;    // '10.0' to '10.9';
    procedure TestAllOps;      // '-+/*()[]{}^**%://'; - changed in version 1.1
    procedure TestExpression;  // '5+8*(7-2)'
    procedure TestEmptyLine1;  // CR
    procedure TestEmptyLine2;  // CR+LF
    procedure TestEmptyLine3;  // '3+'+CR+LF+'5';
    procedure TestUnknown;     // '#a+'
    procedure TestRemovingParenthesisSquare; // added in version 1.1
    procedure TestRemovingParenthesisCurly;  // added in version 1.1
    procedure TestAddingParenthesis;         // added in version 1.1
  end;

  TestNextToken = class(TTestCase)
  strict private
    FTokenizer: TMathTokenizer;
    FSource: string;
    FSources: TStrings;
  protected
    procedure SetSource;
    procedure TestSource;
  public
    procedure SetUp; override;
    procedure TearDown; override;
  published
    procedure TestCreate;
    procedure TestWhiteChar;   // ' '
    procedure TestWhiteString; // '  '
    procedure TestDigit;       // '0' to '9';
    procedure TestWhiteDigit;  // ' 0 ' to ' 9 ';
    procedure TestDoubleDigit; // '10' to '19';
    procedure TestDecimals;    // '10.0' to '10.9';
    procedure TestAllOps;      // '-+/*()^**%://'; - changed in version 1.1
    procedure TestExpression;  // '5+8*(7-2)'
    procedure TestEmptyLine1;  // CR
    procedure TestEmptyLine2;  // CR+LF
    procedure TestEmptyLine3;  // '3+'+CR+LF+'5';
    procedure TestUnknown;     // '#a+'
    procedure TestComment;     // added in version 1.1
  end;

{$IFnDEF NO_NLS}

  { TestEncoding }

  TestEncoding = class(TTestCase)
  strict private
    FTokenizer: TMathTokenizer;
    FSource: string;
  public
    procedure SetUp; override;
    procedure TearDown; override;
  published
    procedure TestEmptyEncoding;
    procedure TestEmptyDecoding;
    procedure TestEncodingNothing;
    procedure TestDecodingNothing;
    procedure TestEncoding;
    procedure TestDecoding;
    procedure TestRoundTrip;
    procedure TestWhenDecimalIsPeriod;
  end;
{$ENDIF}

implementation

procedure TestNextTokenType.SetUp;
begin
  FTokenizer := TMathTokenizer.Create;
  FSource := '';
  FSources := TStringList.Create;
end;

procedure TestNextTokenType.TearDown;
begin
  FTokenizer.Free;
  FTokenizer := nil;
  FSource := '';
  FSources.Free;
  FSources := nil;
end;

procedure TestNextTokenType.SetSource;
var
  i: integer;
begin
  FSource := '';
  for I := 0 to FSources.Count - 1 do
    FSource := FSource + FSources[i];
  FTokenizer.Source := FSource;
end;

procedure TestNextTokenType.TestSource;
var
  i: integer;
  p: integer;
begin
  Check(FTokenizer <> nil, 'FTokenizer broken!');
  CheckEquals(FSource, FTokenizer.Source, 'FTokenizer source');
  p := 1;
  for I := 0 to FSources.Count - 1 do begin
    FTokenizer.NextToken;
    CheckEquals(ptrint(FSources.Objects[i]), ord(FTokenizer.TokenType),
      Format('NextTokenType token="%s" index=%d', [FTokenizer.Token, i]));
    CheckEquals(FSources[i], FTokenizer.Token,
      Format('Token index=%d', [i]));
    CheckEquals(p, FTokenizer.TokenStart,
      Format('TokenStart index=%d', [i]));
    CheckEquals(length(FSources[i]), FTokenizer.TokenLength,
      Format('TokenLength index=%d', [i]));
    p := p + length(FSources[i]);
  end;
  CheckEquals(ord(ttEOS), ord(FTokenizer.NextToken), 'eos');
  CheckEquals(0, FTokenizer.TokenLength, 'eos Tokenlength');
  CheckEquals(p, FTokenizer.TokenStart, 'eos Token start');
end;

procedure TestNextTokenType.TestWhiteChar;
begin
  FSources.AddObject(' ', TObject(ttWhite));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestWhiteString;
begin
  FSources.Clear;
  FSources.AddObject('  ', TObject(ttWhite));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestCreate;
begin
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestDigit;
var
  c: char;
begin
  for c := '0' to '9' do begin
    FSources.Clear;
    FSources.AddObject(c, TObject(ttNumber));
    SetSource;
    TestSource;
  end;
end;

procedure TestNextTokenType.TestWhiteDigit;
var
  c: char;
begin
  for c := '0' to '9' do begin
    FSources.Clear;
    FSources.AddObject(' ', TObject(ttWhite));
    FSources.AddObject(c, TObject(ttNumber));
    FSources.AddObject(' ', TObject(ttWhite));
    SetSource;
    TestSource;
  end;
end;

procedure TestNextTokenType.TestDoubleDigit;
var
  c: char;
begin
  for c := '0' to '9' do begin
    FSources.Clear;
    FSources.AddObject('1' + c, TObject(ttNumber));
    SetSource;
    TestSource;
  end;
end;

procedure TestNextTokenType.TestEmptyLine1;
begin
  FSources.AddObject(#13, TObject(ttWhite));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestEmptyLine2;
begin
  FSources.AddObject(#13#10, TObject(ttWhite));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestEmptyLine3;
begin
  FSources.AddObject('3', TObject(ttNumber));
  FSources.AddObject('+', TObject(ttPlus));
  FSources.AddObject(#13#10, TObject(ttWhite));
  FSources.AddObject('5', TObject(ttNumber));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestExpression;
begin
  // '5+8*(7-2)'
  FSources.AddObject('5', TObject(ttNumber));
  FSources.AddObject('+', TObject(ttPlus));
  FSources.AddObject('8', TObject(ttNumber));
  FSources.AddObject('*', TObject(ttMult));
  FSources.AddObject('(', TObject(ttLeftParenthesis));
  FSources.AddObject('7', TObject(ttNumber));
  FSources.AddObject('-', TObject(ttMinus));
  FSources.AddObject('2', TObject(ttNumber));
  FSources.AddObject(')', TObject(ttRightParenthesis));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestDecimals;
var
  c: char;
begin
  for c := '0' to '9' do begin
    FSources.Clear;
    FSources.AddObject('10' + FTokenizer.DecimalSeparator + c, TObject(ttNumber));
    SetSource;
    TestSource;
  end;
end;

procedure TestNextTokenType.TestAllOps;
begin
  FSources.AddObject('-', TObject(ttMinus));
  FSources.AddObject('+', TObject(ttPlus));
  FSources.AddObject('*', TObject(ttMult));
  FSources.AddObject('/', TObject(ttDiv));
  FSources.AddObject('(', TObject(ttLeftParenthesis));
  FSources.AddObject(')', TObject(ttRightParenthesis));
  FSources.AddObject('[', TObject(ttLeftParenthesis));
  FSources.AddObject(']', TObject(ttRightParenthesis));
  FSources.AddObject('{', TObject(ttLeftParenthesis));
  FSources.AddObject('}', TObject(ttRightParenthesis));
  FSources.AddObject('^', TObject(ttCaret));
  FSources.AddObject('%', TObject(ttPercent));
  FSources.AddObject('**', TObject(ttCaret));
  FSources.AddObject('//comment'#13#10, TObject(ttComment));
  FSources.AddObject(':', TObject(ttColon));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestUnknown;
begin
  FSources.AddObject('#', TObject(ttUnknown));
  FSources.AddObject('a', TObject(ttUnknown));
  FSources.AddObject('+', TObject(ttPlus));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestRemovingParenthesisCurly;
begin
  FTokenizer.DeleteLeftBrace('#');
  FTokenizer.DeleteLeftBrace('(');
  FTokenizer.DeleteLeftBrace('{');
  FSources.AddObject('(', TObject(ttLeftParenthesis));
  FSources.AddObject(')', TObject(ttRightParenthesis));
  FSources.AddObject('[', TObject(ttLeftParenthesis));
  FSources.AddObject(']', TObject(ttRightParenthesis));
  FSources.AddObject('{', TObject(ttUnknown));
  FSources.AddObject('}', TObject(ttUnknown));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestRemovingParenthesisSquare;
begin
  FTokenizer.DeleteLeftBrace('#');
  FTokenizer.DeleteLeftBrace('(');
  FTokenizer.DeleteLeftBrace('[');
  FSources.AddObject('(', TObject(ttLeftParenthesis));
  FSources.AddObject(')', TObject(ttRightParenthesis));
  FSources.AddObject('[', TObject(ttUnknown));
  FSources.AddObject(']', TObject(ttUnknown));
  FSources.AddObject('{', TObject(ttLeftParenthesis));
  FSources.AddObject('}', TObject(ttRightParenthesis));
  SetSource;
  TestSource;
end;

procedure TestNextTokenType.TestAddingParenthesis;
begin
  FTokenizer.DeleteLeftBrace('{');
  FTokenizer.DeleteLeftBrace('[');
  CheckEquals('(', FTokenizer.LeftBraces, '(');
  FTokenizer.AddLeftBrace('{');
  CheckEquals('({', FTokenizer.LeftBraces, '({');
  // check can't add bogus left parent
  FTokenizer.AddLeftBrace('#');
  CheckEquals('({', FTokenizer.LeftBraces, '({ without a');
  FSources.AddObject('(', TObject(ttLeftParenthesis));
  FSources.AddObject(')', TObject(ttRightParenthesis));
  FSources.AddObject('{', TObject(ttLeftParenthesis));
  FSources.AddObject('}', TObject(ttRightParenthesis));
  FSources.AddObject('[', TObject(ttUnknown));
  FSources.AddObject(']', TObject(ttUnknown));
  FSources.AddObject('#', TObject(ttUnknown));
  SetSource;
  TestSource;
end;


{ TestNextToken }

procedure TestNextToken.SetUp;
begin
  FTokenizer := TMathTokenizer.Create;
  FSource := '';
  FSources := TStringList.Create;
end;

procedure TestNextToken.TearDown;
begin
  FTokenizer.Free;
  FTokenizer := nil;
  FSource := '';
  FSources.Free;
  FSources := nil;
end;

procedure TestNextToken.SetSource;
var
  i: integer;
begin
  FSource := '';
  for I := 0 to FSources.Count - 1 do
    FSource := FSource + FSources[i];
  FTokenizer.Source := FSource;
end;

procedure TestNextToken.TestSource;
var
  i: integer;
  p: integer;
  ldw: integer;
  src: string;
begin
  Check(FTokenizer <> nil, 'FTokenizer broken!');
  CheckEquals(FSource, FTokenizer.Source, 'FTokenizer source');
  p := 1;
  for I := 0 to FSources.Count - 1 do begin
    src := FSources[i];
    ldw := 0;
    while (src <> '') and (src[1] <= ' ') do begin
      inc(ldw);
      delete(src, 1, 1);
    end;
    if src <> '' then begin
      FTokenizer.NextTrueToken;
      CheckEquals(ptrint(FSources.Objects[i]), ord(FTokenizer.TokenType),
        Format('NextToken token="%s" index=%d', [FTokenizer.Token, i]));
      CheckEquals(trim(src), FTokenizer.Token,
        Format('Token index=%d', [i]));
      CheckEquals(p+ldw, FTokenizer.TokenStart,
        Format('TokenStart index=%d', [i]));
      CheckEquals(length(trim(src)), FTokenizer.TokenLength,
        Format('TokenLength index=%d', [i]));
    end;
    p := p + length(FSources[i]);
  end;
  CheckEquals(ord(ttEOS), ord(FTokenizer.NextTrueToken), 'eos');
  CheckEquals(0, FTokenizer.TokenLength, 'eos Tokenlength');
  CheckEquals(p, FTokenizer.TokenStart, 'eos Token start');

  src := '';
  FTokenizer.Source := FSource;
  while FTokenizer.NextToken <> ttEOS do
    src := src + FTokenizer.Token;
  CheckEquals(FSource, src, 'reconstructing source');
end;

procedure TestNextToken.TestWhiteChar;
begin
  FSources.AddObject(' ',TObject(ttEos));
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestWhiteString;
begin
  FSources.Clear;
  FSources.AddObject('  ',TObject(ttEos));
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestCreate;
begin
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestDigit;
var
  c: char;
begin
  for c := '0' to '9' do begin
    FSources.Clear;
    FSources.AddObject(c,TObject(ttNumber));
    SetSource;
    TestSource;
  end;
end;

procedure TestNextToken.TestWhiteDigit;
var
  c: char;
begin
  for c := '0' to '9' do begin
    FSources.Clear;
    FSources.AddObject(' ' + c + ' ',TObject(ttNumber));
    SetSource;
    TestSource;
  end;
end;

procedure TestNextToken.TestDoubleDigit;
var
  c: char;
begin
  for c := '0' to '9' do begin
    FSources.Clear;
    FSources.AddObject('  1' + c,TObject(ttNumber));
    SetSource;
    TestSource;
  end;
end;

procedure TestNextToken.TestEmptyLine1;
begin
  FSources.AddObject(#13,TObject(ttEOS));
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestEmptyLine2;
begin
  FSources.AddObject(#13#10,TObject(ttEOS));
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestEmptyLine3;
begin
  FSources.AddObject('3',TObject(ttNumber));
  FSources.AddObject(' +',TObject(ttPlus));
  FSources.AddObject(#13#10' 5',TObject(ttNumber));
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestExpression;
begin
  // '5+8*(7-2)'
  FSources.AddObject(' 5  ',TObject(ttNumber));
  FSources.AddObject('+',TObject(ttPlus));
  FSources.AddObject('8 ',TObject(ttNumber));
  FSources.AddObject(' *',TObject(ttMult));
  FSources.AddObject('( ',TObject(ttLeftParenthesis));
  FSources.AddObject('7',TObject(ttNumber));
  FSources.AddObject(' - ',TObject(ttMinus));
  FSources.AddObject('2',TObject(ttNumber));
  FSources.AddObject('  )',TObject(ttRightParenthesis));
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestDecimals;
var
  c: char;
begin
  for c := '0' to '9' do begin
    FSources.Clear;
    FSources.AddObject('  10' + FTokenizer.DecimalSeparator + c,TObject(ttNumber));
    SetSource;
    TestSource;
  end;
end;

procedure TestNextToken.TestAllOps;
begin
  FSources.AddObject('-',TObject(ttMinus));
  FSources.AddObject('+',TObject(ttPlus));
  FSources.AddObject('*',TObject(ttMult));
  FSources.AddObject('/',TObject(ttDiv));
  FSources.AddObject('(',TObject(ttLeftParenthesis));
  FSources.AddObject(')',TObject(ttRightParenthesis));
  FSources.AddObject('[', TObject(ttLeftParenthesis));
  FSources.AddObject(']', TObject(ttRightParenthesis));
  FSources.AddObject('{', TObject(ttLeftParenthesis));
  FSources.AddObject('}', TObject(ttRightParenthesis));
  FSources.AddObject('^', TObject(ttCaret));
  FSources.AddObject('%', TObject(ttPercent));
  FSources.AddObject('**', TObject(ttCaret));
  FSources.AddObject(':', TObject(ttColon));
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestUnknown;
begin
  FSources.AddObject('#', TObject(ttUnknown));
  FSources.AddObject('a', TObject(ttUnknown));
  FSources.AddObject('+', TObject(ttPlus));
  SetSource;
  TestSource;
end;

procedure TestNextToken.TestComment;
begin
  FTokenizer.Source := '//comment';
  CheckEquals(ord(FTokenizer.NextTrueToken), ord(ttEOS));
end;

{$IFnDEF NO_NLS}

{ TestEncoding }

procedure TestEncoding.SetUp;
begin
  inherited SetUp;
  FTokenizer := TMathTokenizer.Create;
  FTokenizer.DecimalSeparator := ',';
end;

procedure TestEncoding.TearDown;
begin
  FTokenizer.Free;
  inherited TearDown;
end;

procedure TestEncoding.TestEmptyEncoding;
begin
  CheckEquals(0, length(FTokenizer.Encode('')));
end;

procedure TestEncoding.TestEmptyDecoding;
begin
  CheckEquals(0, length(FTokenizer.Decode('')));
end;

procedure TestEncoding.TestEncodingNothing;
begin
  FSource := '  pi : 32 / and.then';
  CheckEquals(FSource, FTokenizer.Encode(FSource));
end;

procedure TestEncoding.TestDecodingNothing;
begin
  FSource := '  pi : 32 / and.then';
  CheckEquals(FSource, FTokenizer.Decode(FSource));
end;

procedure TestEncoding.TestEncoding;
begin
  FSource := '32,157/0,8 +  9898,9';
  CheckEquals('32.157/0.8 +  9898.9', FTokenizer.Encode(FSource));
end;

procedure TestEncoding.TestDecoding;
begin
  FSource := '32.157/0.8 +  9898.9';
  CheckEquals('32,157/0,8 +  9898,9', FTokenizer.Decode(FSource));
end;

procedure TestEncoding.TestRoundTrip;
var
  ts: string;
begin
  FSource := '32,157/0,8 +  9898,9';
  ts := FTokenizer.Encode(FSource);
  CheckFalse(ts = FSource, 'encoding did not change things');
  CheckEquals(FSource, FTokenizer.Decode(ts));
end;

procedure TestEncoding.TestWhenDecimalIsPeriod;
begin
  FTokenizer.DecimalSeparator := '.';
  FSource := '32.157/0.8 +  9898.9';
  CheckEquals(FSource, FTokenizer.Encode(FSource));
  CheckEquals(FSource, FTokenizer.Decode(FSource));
  FSource := '32,157/0,8 +  9898,9';
  CheckEquals(FSource, FTokenizer.Encode(FSource));
  CheckEquals(FSource, FTokenizer.Decode(FSource));
end;

{$ENDIF}

initialization
  RegisterTest(TestNextTokenType);
  RegisterTest(TestNextToken);
  {$IFDEF NLS}
  RegisterTest(TestEncoding);
  {$ENDIF}
end.

